#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Table
# Table management.
# Specification: workspace:v1.0
# Base URL: https://{workspaceId}.{regionId}.xata.sh
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Table(Namespace):

    base_url = "https://{workspaceId}.{regionId}.xata.sh"
    scope = "workspace"

    def createTable(
        self, table_name: str, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Creates a new table with the given name.  Returns 422 if a table with the same name
        already exists.
        Path: /db/{db_branch_name}/tables/{table_name}
        Method: PUT

        :param table_name: str The Table name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}"
        return self.request("PUT", url_path)

    def deleteTable(
        self, table_name: str, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Deletes the table with the given name.
        Path: /db/{db_branch_name}/tables/{table_name}
        Method: DELETE

        :param table_name: str The Table name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}"
        return self.request("DELETE", url_path)

    def updateTable(
        self,
        table_name: str,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        Update table.  Currently there is only one update operation supported: renaming the table
        by providing a new name.  In the example below, we rename a table from “users” to
        “people”:  ```json // PATCH /db/test:main/tables/users  {   "name": "people" } ```
        Path: /db/{db_branch_name}/tables/{table_name}
        Method: PATCH

        :param table_name: str The Table name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}"
        headers = {"content-type": "application/json"}
        return self.request("PATCH", url_path, headers, payload)

    def getTableSchema(
        self, table_name: str, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Get table schema
        Path: /db/{db_branch_name}/tables/{table_name}/schema
        Method: GET

        :param table_name: str The Table name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/schema"
        return self.request("GET", url_path)

    def setTableSchema(
        self,
        table_name: str,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        Update table schema
        Path: /db/{db_branch_name}/tables/{table_name}/schema
        Method: PUT

        :param table_name: str The Table name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/schema"
        headers = {"content-type": "application/json"}
        return self.request("PUT", url_path, headers, payload)

    def getTableColumns(
        self, table_name: str, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Retrieves the list of table columns and their definition.  This endpoint returns the
        column list with object columns being reported with their full dot-separated path
        (flattened).
        Path: /db/{db_branch_name}/tables/{table_name}/columns
        Method: GET

        :param table_name: str The Table name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/columns"
        return self.request("GET", url_path)

    def addTableColumn(
        self,
        table_name: str,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        Adds a new column to the table.  The body of the request should contain the column
        definition.  In the column definition, the 'name' field should contain the full path
        separated by dots.  If the parent objects do not exists, they will be automatically
        created.  For example, passing `"name": "address.city"` will auto-create the `address`
        object if it doesn't exist.
        Path: /db/{db_branch_name}/tables/{table_name}/columns
        Method: POST

        :param table_name: str The Table name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/columns"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def getColumn(
        self,
        table_name: str,
        column_name: str,
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        Get the definition of a single column.  To refer to sub-objects, the column name can
        contain dots.  For example `address.country`.
        Path: /db/{db_branch_name}/tables/{table_name}/columns/{column_name}
        Method: GET

        :param table_name: str The Table name
        :param column_name: str The Column name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/columns/{column_name}"
        return self.request("GET", url_path)

    def deleteColumn(
        self,
        table_name: str,
        column_name: str,
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        Deletes the specified column.  To refer to sub-objects, the column name can contain dots.
        For example `address.country`.
        Path: /db/{db_branch_name}/tables/{table_name}/columns/{column_name}
        Method: DELETE

        :param table_name: str The Table name
        :param column_name: str The Column name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/columns/{column_name}"
        return self.request("DELETE", url_path)

    def updateColumn(
        self,
        table_name: str,
        column_name: str,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        Update column with partial data.  Can be used for renaming the column by providing a new
        "name" field.  To refer to sub-objects, the column name can contain dots.  For example
        `address.country`.
        Path: /db/{db_branch_name}/tables/{table_name}/columns/{column_name}
        Method: PATCH

        :param table_name: str The Table name
        :param column_name: str The Column name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/columns/{column_name}"
        headers = {"content-type": "application/json"}
        return self.request("PATCH", url_path, headers, payload)



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Databases
# Workspace databases management.
# Specification: core:v1.0
# Base URL: https://api.xata.io
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Databases(Namespace):

    base_url = "https://api.xata.io"
    scope = "core"

    def getDatabaseList(self, workspace_id: str) -> Response:
        """
        List all databases available in your Workspace.
        Path: /workspaces/{workspace_id}/dbs
        Method: GET

        :param workspace_id: str Workspace ID

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/dbs"
        return self.request("GET", url_path)

    def getDatabaseMetadata(self, workspace_id: str, db_name: str) -> Response:
        """
        Retrieve metadata of the given database
        Path: /workspaces/{workspace_id}/dbs/{db_name}
        Method: GET

        :param workspace_id: str Workspace ID
        :param db_name: str The Database Name

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/dbs/{db_name}"
        return self.request("GET", url_path)

    def createDatabase(
        self, workspace_id: str, db_name: str, payload: dict
    ) -> Response:
        """
        Create Database with identifier name
        Path: /workspaces/{workspace_id}/dbs/{db_name}
        Method: PUT

        :param workspace_id: str Workspace ID
        :param db_name: str The Database Name
        :param payload: dict content

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/dbs/{db_name}"
        headers = {"content-type": "application/json"}
        return self.request("PUT", url_path, headers, payload)

    def deleteDatabase(self, workspace_id: str, db_name: str) -> Response:
        """
        Delete a database and all of its branches and tables permanently.
        Path: /workspaces/{workspace_id}/dbs/{db_name}
        Method: DELETE

        :param workspace_id: str Workspace ID
        :param db_name: str The Database Name

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/dbs/{db_name}"
        return self.request("DELETE", url_path)

    def updateDatabaseMetadata(
        self, workspace_id: str, db_name: str, payload: dict
    ) -> Response:
        """
        Update the color of the selected database
        Path: /workspaces/{workspace_id}/dbs/{db_name}
        Method: PATCH

        :param workspace_id: str Workspace ID
        :param db_name: str The Database Name
        :param payload: dict content

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/dbs/{db_name}"
        headers = {"content-type": "application/json"}
        return self.request("PATCH", url_path, headers, payload)

    def listRegions(self, workspace_id: str) -> Response:
        """
        List regions available to create a database on
        Path: /workspaces/{workspace_id}/regions
        Method: GET

        :param workspace_id: str Workspace ID

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/regions"
        return self.request("GET", url_path)



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Users
# Users management
# Specification: core:v1.0
# Base URL: https://api.xata.io
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Users(Namespace):

    base_url = "https://api.xata.io"
    scope = "core"

    def getUser(
        self,
    ) -> Response:
        """
        Return details of the user making the request
        Path: /user
        Method: GET


        :return Response
        """
        url_path = "/user"
        return self.request("GET", url_path)

    def updateUser(self, payload: dict) -> Response:
        """
        Update user info
        Path: /user
        Method: PUT

        :param payload: dict content

        :return Response
        """
        url_path = "/user"
        headers = {"content-type": "application/json"}
        return self.request("PUT", url_path, headers, payload)

    def deleteUser(
        self,
    ) -> Response:
        """
        Delete the user making the request
        Path: /user
        Method: DELETE


        :return Response
        """
        url_path = "/user"
        return self.request("DELETE", url_path)



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Workspaces
# Workspaces management
# Specification: core:v1.0
# Base URL: https://api.xata.io
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Workspaces(Namespace):

    base_url = "https://api.xata.io"
    scope = "core"

    def getWorkspacesList(
        self,
    ) -> Response:
        """
        Retrieve the list of workspaces the user belongs to
        Path: /workspaces
        Method: GET


        :return Response
        """
        url_path = "/workspaces"
        return self.request("GET", url_path)

    def createWorkspace(self, payload: dict) -> Response:
        """
        Creates a new workspace with the user requesting it as its single owner.
        Path: /workspaces
        Method: POST

        :param payload: dict content

        :return Response
        """
        url_path = "/workspaces"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def getWorkspace(self, workspace_id: str) -> Response:
        """
        Retrieve workspace info from a workspace ID
        Path: /workspaces/{workspace_id}
        Method: GET

        :param workspace_id: str Workspace ID

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}"
        return self.request("GET", url_path)

    def updateWorkspace(self, workspace_id: str, payload: dict) -> Response:
        """
        Update workspace info
        Path: /workspaces/{workspace_id}
        Method: PUT

        :param workspace_id: str Workspace ID
        :param payload: dict content

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}"
        headers = {"content-type": "application/json"}
        return self.request("PUT", url_path, headers, payload)

    def deleteWorkspace(self, workspace_id: str) -> Response:
        """
        Delete the workspace with the provided ID
        Path: /workspaces/{workspace_id}
        Method: DELETE

        :param workspace_id: str Workspace ID

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}"
        return self.request("DELETE", url_path)

    def getWorkspaceMembersList(self, workspace_id: str) -> Response:
        """
        Retrieve the list of members of the given workspace
        Path: /workspaces/{workspace_id}/members
        Method: GET

        :param workspace_id: str Workspace ID

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/members"
        return self.request("GET", url_path)

    def updateWorkspaceMemberRole(
        self, workspace_id: str, user_id: str, payload: dict
    ) -> Response:
        """
        Update a workspace member role.  Workspaces must always have at least one owner, so this
        operation will fail if trying to remove owner role from the last owner in the workspace.
        Path: /workspaces/{workspace_id}/members/{user_id}
        Method: PUT

        :param workspace_id: str Workspace ID
        :param user_id: str UserID
        :param payload: dict content

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/members/{user_id}"
        headers = {"content-type": "application/json"}
        return self.request("PUT", url_path, headers, payload)

    def removeWorkspaceMember(self, workspace_id: str, user_id: str) -> Response:
        """
        Remove the member from the workspace
        Path: /workspaces/{workspace_id}/members/{user_id}
        Method: DELETE

        :param workspace_id: str Workspace ID
        :param user_id: str UserID

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/members/{user_id}"
        return self.request("DELETE", url_path)



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Invites
# Manage user invites.
# Specification: core:v1.0
# Base URL: https://api.xata.io
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Invites(Namespace):

    base_url = "https://api.xata.io"
    scope = "core"

    def inviteWorkspaceMember(self, workspace_id: str, payload: dict) -> Response:
        """
        Invite some user to join the workspace with the given role
        Path: /workspaces/{workspace_id}/invites
        Method: POST

        :param workspace_id: str Workspace ID
        :param payload: dict content

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/invites"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def cancelWorkspaceMemberInvite(
        self, workspace_id: str, invite_id: str
    ) -> Response:
        """
        This operation provides a way to cancel invites by deleting them.  Already accepted
        invites cannot be deleted.
        Path: /workspaces/{workspace_id}/invites/{invite_id}
        Method: DELETE

        :param workspace_id: str Workspace ID
        :param invite_id: str Invite identifier

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/invites/{invite_id}"
        return self.request("DELETE", url_path)

    def updateWorkspaceMemberInvite(
        self, workspace_id: str, invite_id: str, payload: dict
    ) -> Response:
        """
        This operation provides a way to update an existing invite.  Updates are performed in-
        place; they do not change the invite link, the expiry time, nor do they re-notify the
        recipient of the invite.
        Path: /workspaces/{workspace_id}/invites/{invite_id}
        Method: PATCH

        :param workspace_id: str Workspace ID
        :param invite_id: str Invite identifier
        :param payload: dict content

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/invites/{invite_id}"
        headers = {"content-type": "application/json"}
        return self.request("PATCH", url_path, headers, payload)

    def acceptWorkspaceMemberInvite(
        self, workspace_id: str, invite_key: str
    ) -> Response:
        """
        Accept the invitation to join a workspace.  If the operation succeeds the user will be a
        member of the workspace
        Path: /workspaces/{workspace_id}/invites/{invite_key}/accept
        Method: POST

        :param workspace_id: str Workspace ID
        :param invite_key: str Invite Key (secret) for the invited user

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/invites/{invite_key}/accept"
        return self.request("POST", url_path)

    def resendWorkspaceMemberInvite(
        self, workspace_id: str, invite_id: str
    ) -> Response:
        """
        This operation provides a way to resend an Invite notification.  Invite notifications can
        only be sent for Invites not yet accepted.
        Path: /workspaces/{workspace_id}/invites/{invite_id}/resend
        Method: POST

        :param workspace_id: str Workspace ID
        :param invite_id: str Invite identifier

        :return Response
        """
        url_path = f"/workspaces/{workspace_id}/invites/{invite_id}/resend"
        return self.request("POST", url_path)



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Search_and_filter
# APIs for searching, querying, filtering, and aggregating records.
# Specification: workspace:v1.0
# Base URL: https://{workspaceId}.{regionId}.xata.sh
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Search_and_filter(Namespace):

    base_url = "https://{workspaceId}.{regionId}.xata.sh"
    scope = "workspace"

    def queryTable(
        self,
        table_name: str,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        The Query Table API can be used to retrieve all records in a table.  The API support
        filtering, sorting, selecting a subset of columns, and pagination.  The overall structure
        of the request looks like this:  ```json // POST
        /db/<dbname>:<branch>/tables/<table>/query {   "columns": [...],   "filter": {     "$all":
        [...],     "$any": [...]     ...   },   "sort": {     "multiple": [...]     ...   },
        "page": {     ...   } } ```  For usage, see also the [API Guide](https://xata.io/docs/api-
        guide/get).  ### Column selection  If the `columns` array is not specified, all columns
        are included.  For link fields, only the ID column of the linked records is included in
        the response.  If the `columns` array is specified, only the selected and internal columns
        `id` and `xata` are included.  The `*` wildcard can be used to select all columns.  For
        objects and link fields, if the column name of the object is specified, we include all of
        its sub-keys.  If only some sub-keys are specified (via dotted notation, e.g.
        `"settings.plan"` ), then only those sub-keys from the object are included.  By the way of
        example, assuming two tables like this:  ```json {"truncate": true} {   "tables": [     {
        "name": "teams",       "columns": [         {           "name": "name",           "type":
        "string"         },         {           "name": "owner",           "type": "link",
        "link": {             "table": "users"           }         },         {           "name":
        "foundedDate",           "type": "datetime"         },       ]     },     {       "name":
        "users",       "columns": [         {           "name": "email",           "type": "email"
        },         {           "name": "full_name",           "type": "string"         },
        {           "name": "address",           "type": "object",           "columns": [
        {               "name": "street",               "type": "string"             },
        {               "name": "number",               "type": "int"             },             {
        "name": "zipcode",               "type": "int"             }           ]         },
        {           "name": "team",           "type": "link",           "link": {
        "table": "teams"           }         }       ]     }   ] } ```  A query like this:
        ```json POST /db/<dbname>:<branch>/tables/<table>/query {   "columns": [     "name",
        "address.*"   ] } ```  returns objects like:  ```json {   "name": "Kilian",   "address": {
        "street": "New street",     "number": 41,     "zipcode": 10407   } } ```  while a query
        like this:  ```json POST /db/<dbname>:<branch>/tables/<table>/query {   "columns": [
        "name",     "address.street"   ] } ```  returns objects like:  ```json {   "id": "id1"
        "xata": {     "version": 0   }   "name": "Kilian",   "address": {     "street": "New
        street"   } } ```  If you want to return all columns from the main table and selected
        columns from the linked table, you can do it like this:  ```json {   "columns": ["*",
        "team.name"] } ```  The `"*"` in the above means all columns, including columns of
        objects.  This returns data like:  ```json {   "id": "id1"   "xata": {     "version": 0
        }   "name": "Kilian",   "email": "kilian@gmail.com",   "address": {     "street": "New
        street",     "number": 41,     "zipcode": 10407   },   "team": {     "id": "XX",
        "xata": {       "version": 0     },     "name": "first team"   } } ```  If you want all
        columns of the linked table, you can do:  ```json {   "columns": ["*", "team.*"] } ```
        This returns, for example:  ```json {   "id": "id1"   "xata": {     "version": 0   }
        "name": "Kilian",   "email": "kilian@gmail.com",   "address": {     "street": "New
        street",     "number": 41,     "zipcode": 10407   },   "team": {     "id": "XX",
        "xata": {       "version": 0     },     "name": "first team",     "code": "A1",
        "foundedDate": "2020-03-04T10:43:54.32Z"   } } ```  ### Filtering  There are two types of
        operators:  - Operators that work on a single column: `$is`, `$contains`, `$pattern`,
        `$includes`, `$gt`, etc.  - Control operators that combine multiple conditions: `$any`,
        `$all`, `$not` ,   `$none`, etc.  All operators start with an `$` to differentiate them
        from column names (which are not allowed to start with a dollar sign).  #### Exact
        matching and control operators  Filter by one column:  ```json {   "filter": {
        "<column_name>": "value"   } } ```  This is equivalent to using the `$is` operator:
        ```json {   "filter": {     "<column_name>": {       "$is": "value"     }   } } ```  For
        example:  ```json {   "filter": {     "name": "r2"   } } ```  Or:  ```json {   "filter": {
        "name": {       "$is": "r2"     }   } } ```  For objects, both dots and nested versions
        work:  ```json {   "filter": {     "settings.plan": "free"   } } ```  ```json {
        "filter": {     "settings": {       "plan": "free"     }   } } ```  If you want to OR
        together multiple values, you can use the `$any` operator with an array of values:
        ```json {   "filter": {     "settings.plan": { "$any": ["free", "paid"] }   } } ```  If
        you specify multiple columns in the same filter, they are logically AND'ed together:
        ```json {   "filter": {     "settings.dark": true,     "settings.plan": "free"   } } ```
        The above matches if both conditions are met.  To be more explicit about it, you can use
        `$all` or `$any`:  ```json {   "filter": {     "$any": {       "settings.dark": true,
        "settings.plan": "free"     }   } } ```  The `$all` and `$any` operators can also receive
        an array of objects, which allows for repeating column names:  ```json {   "filter": {
        "$any": [       {         "name": "r1"       },       {         "name": "r2"       }     ]
        } } ```  You can check for a value being not-null with `$exists`:  ```json {   "filter": {
        "$exists": "settings"   } } ```  This can be combined with `$all` or `$any` :  ```json {
        "filter": {     "$all": [       {         "$exists": "settings"       },       {
        "$exists": "name"       }     ]   } } ```  Or you can use the inverse operator
        `$notExists`:  ```json {   "filter": {     "$notExists": "settings"   } } ```  ####
        Partial match  `$contains` is the simplest operator for partial matching.  Note that
        `$contains` operator can cause performance issues at scale, because indices cannot be
        used.  ```json {   "filter": {     "<column_name>": {       "$contains": "value"     }   }
        } ```  Wildcards are supported via the `$pattern` operator:  ```json {   "filter": {
        "<column_name>": {       "$pattern": "v*alu?"     }   } } ```  The `$pattern` operator
        accepts two wildcard characters: * `*` matches zero or more characters * `?` matches
        exactly one character  If you want to match a string that contains a wildcard character,
        you can escape them using a backslash (`\`). You can escape a backslash by usign another
        backslash.  You can also use the `$endsWith` and `$startsWith` operators:  ```json {
        "filter": {     "<column_name>": {       "$endsWith": ".gz"     },     "<column_name>": {
        "$startsWith": "tmp-"     }   } } ```  #### Numeric or datetime ranges  ```json {
        "filter": {     "<column_name>": {       "$ge": 0,       "$lt": 100     }   } } ``` Date
        ranges support the same operators, with the date using the format defined in [RFC
        3339](https://www.rfc-editor.org/rfc/rfc3339): ```json {   "filter": {
        "<column_name>": {       "$gt": "2019-10-12T07:20:50.52Z",       "$lt":
        "2021-10-12T07:20:50.52Z"     }   } } ``` The supported operators are `$gt`, `$lt`, `$ge`,
        `$le`.  #### Negations  A general `$not` operator can inverse any operation.  ```json {
        "filter": {     "$not": {       "<column_name1>": "value1",       "<column_name2>":
        "value1"     }   } } ```  Note: in the above the two condition are AND together, so this
        does (NOT ( ... AND ...))  Or more complex:  ```json {   "filter": {     "$not": {
        "$any": [         {           "<column_name1>": "value1"         },         {
        "$all": [             {               "<column_name2>": "value2"             },
        {               "<column_name3>": "value3"             }           ]         }       ]
        }   } } ```  The `$not: { $any: {}}` can be shorted using the `$none` operator:  ```json {
        "filter": {     "$none": {       "<column_name1>": "value1",       "<column_name2>":
        "value1"     }   } } ```  In addition, you can use operators like `$isNot` or `$notExists`
        to simplify expressions:  ```json {   "filter": {     "<column_name>": {       "$isNot":
        "2019-10-12T07:20:50.52Z"     }   } } ```  #### Working with arrays  To test that an array
        contains a value, use `$includesAny`.  ```json {   "filter": {     "<array_name>": {
        "$includesAny": "value"     }   } } ```  ##### `includesAny`  The `$includesAny` operator
        accepts a custom predicate that will check if any value in the array column matches the
        predicate.  The `$includes` operator is a synonym for the `$includesAny` operator.  For
        example a complex predicate can include the `$all` , `$contains` and `$endsWith`
        operators:  ```json {   "filter": {     "<array name>": {       "$includes": {
        "$all": [           { "$contains": "label" },           { "$not": { "$endsWith": "-debug"
        } }         ]       }     }   } } ```  ##### `includesNone`  The `$includesNone` operator
        succeeds if no array item matches the predicate.  ```json {   "filter": {
        "settings.labels": {       "$includesNone": [{ "$contains": "label" }]     }   } } ``` The
        above matches if none of the array values contain the string "label".  ##### `includesAll`
        The `$includesAll` operator succeeds if all array items match the predicate.  Here is an
        example of using the `$includesAll` operator:  ```json {   "filter": {
        "settings.labels": {       "$includesAll": [{ "$contains": "label" }]     }   } } ```  The
        above matches if all array values contain the string "label".  ### Sorting  Sorting by one
        element:  ```json POST /db/demo:main/tables/table/query {   "sort": {     "index": "asc"
        } } ```  or descendently:  ```json POST /db/demo:main/tables/table/query {   "sort": {
        "index": "desc"   } } ```  Sorting by multiple fields:  ```json POST
        /db/demo:main/tables/table/query {   "sort": [     {       "index": "desc"     },     {
        "createdAt": "desc"     }   ] } ```  ### Pagination  We offer cursor pagination and offset
        pagination.  For queries that are expected to return more than 1000 records, cursor
        pagination is needed in order to retrieve all of their results.  The offset pagination
        method is limited to 1000 records.  Example of size + offset pagination:  ```json POST
        /db/demo:main/tables/table/query {   "page": {     "size": 100,     "offset": 200   } }
        ```  The `page.size` parameter represents the maximum number of records returned by this
        query.  It has a default value of 20 and a maximum value of 200. The `page.offset`
        parameter represents the number of matching records to skip.  It has a default value of 0
        and a maximum value of 800.  Example of cursor pagination:  ```json POST
        /db/demo:main/tables/table/query {   "page": {
        "after":"fMoxCsIwFIDh3WP8c4amDai5hO5SJCRNfaVSeC9b6d1FD"   } } ```  In the above example,
        the value of the `page.after` parameter is the cursor returned by the previous query.  A
        sample response is shown below:  ```json {   "meta": {     "page": {       "cursor":
        "fMoxCsIwFIDh3WP8c4amDai5hO5SJCRNfaVSeC9b6d1FD",       "more": true     }   },
        "records": [...] } ```  The `page` object might contain the follow keys, in addition to
        `size` and `offset` that were introduced before:  - `after`: Return the next page 'after'
        the current cursor - `before`: Return the previous page 'before' the current cursor.  -
        `start`: Resets the given cursor position to the beginning of the query result set.  Will
        return the first N records from the query result, where N is the `page.size` parameter.  -
        `end`: Resets the give cursor position to the end for the query result set.  Returns the
        last N records from the query result, where N is the `page.size` parameter.  The request
        will fail if an invalid cursor value is given to `page.before`, `page.after`, `page.start`
        , or `page.end`. No other cursor setting can be used if `page.start` or `page.end` is set
        in a query.  If both `page.before` and `page.after` parameters are present we treat the
        request as a range query.  The range query will return all entries after `page.after`, but
        before `page.before`, up to `page.size` or the maximum page size.  This query requires
        both cursors to use the same filters and sort settings, plus we require `page.after <
        page.before`. The range query returns a new cursor.  If the range encompass multiple pages
        the next page in the range can be queried by update `page.after` to the returned cursor
        while keeping the `page.before` cursor from the first range query.  The `filter` ,
        `columns`, `sort` , and `page.size` configuration will be encoded with the cursor.  The
        pagination request will be invalid if `filter` or `sort` is set.  The columns returned and
        page size can be changed anytime by passing the `columns` or `page.size` settings to the
        next query.  **Special cursors:**  - `page.after=end`: Result points past the last entry.
        The list of records   returned is empty, but `page.meta.cursor` will include a cursor that
        can be   used to "tail" the table from the end waiting for new data to be inserted.  -
        `page.before=end`: This cursor returns the last page.  - `page.start=<cursor>`: Start at
        the beginning of the result set of the <cursor> query.  This is equivalent to querying the
        first page without a cursor but applying `filter` and `sort` . Yet the `page.start`
        cursor can be convenient at times as user code does not need to remember the   filter,
        sort, columns or page size configuration.  All these information are   read from the
        cursor.  - `page.end=<cursor>`: Move to the end of the result set of the <cursor> query.
        This is equivalent to querying the   last page with `page.before=end`, `filter`, and
        `sort` . Yet the   `page.end` cursor can be more convenient at times as user code does not
        need to remember the filter, sort, columns or page size configuration.  All   these
        information are read from the cursor.  When using special cursors like `page.after="end"`
        or `page.before="end"`, we still allow `filter` and `sort` to be set.  Example of getting
        the last page:  ```json POST /db/demo:main/tables/table/query {   "page": {     "size":
        10,     "before": "end"   } } ```
        Path: /db/{db_branch_name}/tables/{table_name}/query
        Method: POST

        :param table_name: str The Table name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/query"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def searchBranch(
        self, payload: dict, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Run a free text search operation across the database branch.
        Path: /db/{db_branch_name}/search
        Method: POST

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/search"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def searchTable(
        self,
        table_name: str,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        Run a free text search operation in a particular table.  The endpoint accepts a `query`
        parameter that is used for the free text search and a set of structured filters (via the
        `filter` parameter) that are applied before the search.  The `filter` parameter uses the
        same syntax as the [query endpoint](/api-reference/db/db_branch_name/tables/table_name/)
        with the following exceptions: * filters `$contains`, `$startsWith`, `$endsWith` don't
        work on columns of type `text` * filtering on columns of type `multiple` is currently
        unsupported
        Path: /db/{db_branch_name}/tables/{table_name}/search
        Method: POST

        :param table_name: str The Table name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/search"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def summarizeTable(
        self,
        table_name: str,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        This endpoint allows you to (optionally) define groups, and then to run calculations on
        the values in each group.  This is most helpful when  you'd like to understand the data
        you have in your database.  A group is a combination of unique values.  If you create a
        group for  `sold_by`, `product_name`, we will return one row for every combination  of
        `sold_by` and `product_name` you have in your database.  When you  want to calculate
        statistics, you define these groups and ask Xata to  calculate data on each group.  **Some
        questions you can ask of your data:**  How many records do I have in this table?  - Set
        `columns: []` as we we want data from the entire table, so we ask for no groups.  - Set
        `summaries: {"total": {"count": "*"}}` in order to see the count  of all records.  We use
        `count: *` here we'd like to know the total  amount of rows; ignoring whether they are
        `null` or not.  What are the top total sales for each product in July 2022 and sold  more
        than 10 units?  - Set `filter: {soldAt: {   "$ge": "2022-07-01T00:00:00.000Z",    "$lt":
        "2022-08-01T00:00:00.000Z"} }`  in order to limit the result set to sales recorded in July
        2022. - Set `columns: [product_name]` as we'd like to run calculations on  each unique
        product name in our table.  Setting `columns` like this will  produce one row per unique
        product name.  - Set `summaries: {"total_sales": {"count": "product_name"}}` as we'd  like
        to create a field called "total_sales" for each group.  This field  will count all rows in
        each group with non-null product names.  - Set `sort: [{"total_sales": "desc"}]` in order
        to bring the rows with  the highest total_sales field to the top.  - Set `summariesFilter:
        {"total_sales": {"$ge": 10}}` to only send back data  with greater than or equal to 10
        units.  `columns`: tells Xata how to create each group.  If you add `product_id`  we will
        create a new group for every unique `product_id`.  `summaries`: tells Xata which
        calculations to run on each group.  Xata currently supports count, min, max, sum, average.
        `sort`: tells Xata in which order you'd like to see results.  You may  sort by fields
        specified in `columns` as well as the summary names  defined in `summaries`.  note:
        Sorting on summarized values can be slower on very large tables;  this will impact your
        rate limit significantly more than other queries.  Try use `filter` to reduce the amount
        of data being processed in order  to reduce impact on your limits.  `summariesFilter`:
        tells Xata how to filter the results of a summary.  It has the same syntax as `filter`,
        however, by using `summariesFilter` you may also filter on the results of a query.  note:
        This is a much slower to use than `filter`. We recommend using  `filter` wherever possible
        and `summariesFilter` when it's not  possible to use `filter`.  `page.size`: tells Xata
        how many records to return.  If unspecified, Xata will return the default size.
        Path: /db/{db_branch_name}/tables/{table_name}/summarize
        Method: POST

        :param table_name: str The Table name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/summarize"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def aggregateTable(
        self,
        table_name: str,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        This endpoint allows you to run aggragations (analytics) on the data from one table.
        While the summary endpoint is served from a transactional store and the results are
        strongly  consistent, the aggregate endpoint is served from our columnar store and the
        results are  only eventually consistent.  On the other hand, the aggregate endpoint uses a
        store that is more appropiate for analytics, makes use of approximative algorithms  (e.g
        for cardinality), and is generally faster and can do more complex aggregations.  For
        usage, see the [API Guide](https://xata.io/docs/api-guide/aggregate).
        Path: /db/{db_branch_name}/tables/{table_name}/aggregate
        Method: POST

        :param table_name: str The Table name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/aggregate"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Authentication
# Authentication and API Key management.
# Specification: core:v1.0
# Base URL: https://api.xata.io
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Authentication(Namespace):

    base_url = "https://api.xata.io"
    scope = "core"

    def getUserAPIKeys(
        self,
    ) -> Response:
        """
        Retrieve a list of existing user API keys
        Path: /user/keys
        Method: GET


        :return Response
        """
        url_path = "/user/keys"
        return self.request("GET", url_path)

    def createUserAPIKey(self, key_name: str) -> Response:
        """
        Create and return new API key
        Path: /user/keys/{key_name}
        Method: POST

        :param key_name: str API Key name

        :return Response
        """
        url_path = f"/user/keys/{key_name}"
        return self.request("POST", url_path)

    def deleteUserAPIKey(self, key_name: str) -> Response:
        """
        Delete an existing API key
        Path: /user/keys/{key_name}
        Method: DELETE

        :param key_name: str API Key name

        :return Response
        """
        url_path = f"/user/keys/{key_name}"
        return self.request("DELETE", url_path)



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#


class UnauthorizedException(Exception):
    pass


class RateLimitException(Exception):
    pass


class RecordNotFoundException(Exception):
    id: str

    def __init__(self, id: str):
        self.id = id

    def __str__(self) -> str:
        return f"Record with id {self.id} not found"


class BadRequestException(Exception):
    status_code: int
    message: str

    def __init__(self, status_code, message):
        self.status_code = status_code
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"Bad request: {self.status_code} {self.message}"


class ServerErrorException(Exception):
    status_code: int
    message: str

    def __init__(self, status_code, message):
        self.status_code = status_code
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"Server error: {self.status_code} {self.message}"



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import logging
import time
from threading import Lock, Thread

from .client import XataClient

DEFAULT_THREAD_POOL_SIZE = 4
DEFAULT_BATCH_SIZE = 25
DEFAULT_FLUSH_INTERVAL = 5
DEFAULT_PROCESSING_TIMEOUT = 0.025


class BulkProcessor(object):
    """
    !!! This helper is still work in progress !!!

    Additional abstraction for bulk requests that process'
    requests in parallel
    :stability alpha
    """

    def __init__(
        self,
        client: XataClient,
        thread_pool_size: int = DEFAULT_THREAD_POOL_SIZE,
        batch_size: int = DEFAULT_BATCH_SIZE,
        flush_interval: int = DEFAULT_FLUSH_INTERVAL,
        processing_timeout: float = DEFAULT_PROCESSING_TIMEOUT,
    ):
        """
        BulkProcessor: Abstraction for bulk ingestion of records.

        :stability alpha


        :param client: XataClient
        :param thread_pool_size: int
        :param batch_size: int
        :param flush_interval: int
        :processing_timeout: float
        """
        if thread_pool_size < 1:
            raise Exception(
                "thread pool size must be greater than 0, default: %d"
                % DEFAULT_THREAD_POOL_SIZE
            )
        if processing_timeout < 0:
            raise Exception(
                "processing timeout can not be negative, default: %f"
                % DEFAULT_PROCESSING_TIMEOUT
            )

        self.client = client
        self.processing_timeout = processing_timeout
        self.batch_size = batch_size
        self.flush_interval = flush_interval
        self.stats = {"total": 0, "queue": 0, "tables": {}}
        self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")

        self.thread_workers = []
        self.records = self.Records(self.batch_size, self.flush_interval, self.logger)

        for i in range(thread_pool_size):
            worker = Thread(
                target=self.process, daemon=True, args=(i,), name="worker-%d" % i
            )
            worker.start()
            self.thread_workers.append(worker)

    def process(self, id: int):
        """
        Process the records
        """
        self.logger.debug(
            "thread #%d: starting bulk processor [thread_pool_size=%d, batch_size=%d, flush_interval=%d, processing_timeout=%s"
            % (
                id,
                len(self.thread_workers),
                self.batch_size,
                self.flush_interval,
                self.processing_timeout,
            )
        )
        while True:
            batch = self.records.next_batch()
            if "table" in batch and len(batch["records"]) > 0:
                r = self.client.records().bulkInsertTableRecords(
                    batch["table"], {"records": batch["records"]}
                )
                if r.status_code != 200:
                    self.logger.error(
                        "thread #%d: unable to process batch for table '%s', with error: %d - %s"
                        % (id, batch["table"], r.status_code, r.json()["message"])
                    )
                    # TODO add records to batch again or callback
                    raise Exception(r.json()["message"])

                self.logger.debug(
                    "thread #%d: pushed a batch of %d records to table %s"
                    % (id, len(batch["records"]), batch["table"])
                )
                self.stats["total"] += len(batch["records"])
                self.stats["queue"] = self.records.size()
                if batch["table"] not in self.stats["tables"]:
                    self.stats["tables"][batch["table"]] = 0
                self.stats["tables"][batch["table"]] += len(batch["records"])
            time.sleep(self.processing_timeout)

    def put_record(self, table_name: str, record: dict):
        """
        Put a record to the processing queue

        :param table_name: str
        :param record: dict
        """
        self.records.put(table_name, [record])

    def put_records(self, table_name: str, records: list[dict]):
        """
        Put multtiple records to the processing queue

        :param table_name: str
        :param records: list[dict]
        """
        self.records.put(table_name, records)

    def dlq(self):
        """
        TODO
        """
        pass

    def get_stats(self):
        """
        Get processing statistics

        :return dict
        """
        return self.stats

    def flush_queue(self):
        """
        Flush all records from the queue.
        TODO: this method does not work correctly yet. The queue is fully flushed
        """
        self.logger.debug("flushing queue with %d records .." % (self.records.size()))
        self.records.set_flush_interval(0)
        self.processing_timeout = 0
        while self.stats["queue"] > 0:
            time.sleep(self.processing_timeout / len(self.thread_workers) + 0.01)

    class Records(object):
        """
        Thread safe storage for records to persist by the bulk processor
        """

        def __init__(self, batch_size: int, flush_interval: int, logger):
            """
            :param batch_size: int
            :param flush_interval: int
            """
            self.batch_size = batch_size
            self.flush_interval = flush_interval
            self.logger = logger

            self.store = dict()
            self.store_ptr = 0
            self.lock = Lock()

        def set_flush_interval(self, interval: int):
            self.flush_interval = interval

        def put(self, table_name: str, records: list[dict]):
            """
            :param table_name: str
            :param records: list[dict]
            """
            with self.lock:
                if table_name not in self.store.keys():
                    self.store[table_name] = {
                        "lock": Lock(),
                        "flushed": time.time(),
                        "records": list(),
                    }
            with self.store[table_name]["lock"]:
                self.store[table_name]["records"] += records

        def next_batch(self) -> dict:
            """
            Get the next batch of records to persist

            :return dict
            """
            table_name = ""
            with self.lock:
                names = list(self.store.keys())
                if len(names) == 0:
                    return {}

                self.store_ptr += 1
                if len(names) <= self.store_ptr:
                    self.store_ptr = 0
                table_name = names[self.store_ptr]

            rs = []
            with self.store[table_name]["lock"]:
                # flush interval exceeded
                time_elapsed = time.time() - self.store[table_name]["flushed"]
                flush_needed = time_elapsed > self.flush_interval
                if flush_needed and len(self.store[table_name]["records"]) > 0:
                    self.logger.debug(
                        "flushing table '%s' with %d records after interval %s > %d"
                        % (
                            table_name,
                            len(self.store[table_name]["records"]),
                            time_elapsed,
                            self.flush_interval,
                        )
                    )
                # pop records ?
                if (
                    len(self.store[table_name]["records"]) >= self.batch_size
                    or flush_needed
                ):
                    self.store[table_name]["flushed"] = time.time()
                    rs = self.store[table_name]["records"][0 : self.batch_size]
                    del self.store[table_name]["records"][0 : self.batch_size]
            return {"table": table_name, "records": rs}

        def length(self, table_name: str) -> int:
            """
            Get record count of a table

            :param table_name: str
            """
            with self.store[table_name]["lock"]:
                return len(self.store[table_name]["records"])

        def size(self) -> int:
            """
            Get total size of stored records
            """
            with self.lock:
                return sum([len(self.store[n]["records"]) for n in self.store.keys()])



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Records
# Record access API.
# Specification: workspace:v1.0
# Base URL: https://{workspaceId}.{regionId}.xata.sh
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Records(Namespace):

    base_url = "https://{workspaceId}.{regionId}.xata.sh"
    scope = "workspace"

    def branchTransaction(
        self, payload: dict, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Execute a transaction on a branch
        Path: /db/{db_branch_name}/transaction
        Method: POST

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/transaction"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def insertRecord(
        self,
        table_name: str,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
        columns: list = None,
    ) -> Response:
        """
        Insert a new Record into the Table
        Path: /db/{db_branch_name}/tables/{table_name}/data
        Method: POST

        :param table_name: str The Table name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.
        :param columns: list = None Column filters

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/data"
        if columns is not None:
            url_path += "?columns=%s" % ",".join(columns)
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def getRecord(
        self,
        table_name: str,
        record_id: str,
        db_name: str = None,
        branch_name: str = None,
        columns: list = None,
    ) -> Response:
        """
        Retrieve record by ID
        Path: /db/{db_branch_name}/tables/{table_name}/data/{record_id}
        Method: GET

        :param table_name: str The Table name
        :param record_id: str The Record name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.
        :param columns: list = None Column filters

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/data/{record_id}"
        if columns is not None:
            url_path += "?columns=%s" % ",".join(columns)
        return self.request("GET", url_path)

    def insertRecordWithID(
        self,
        table_name: str,
        record_id: str,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
        columns: list = None,
        createOnly: bool = None,
        ifVersion: int = None,
    ) -> Response:
        """
        By default, IDs are auto-generated when data is insterted into Xata.  Sending a request to
        this endpoint allows us to insert a record with a pre-existing ID, bypassing the default
        automatic ID generation.
        Path: /db/{db_branch_name}/tables/{table_name}/data/{record_id}
        Method: PUT

        :param table_name: str The Table name
        :param record_id: str The Record name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.
        :param columns: list = None Column filters
        :param createOnly: bool = None
        :param ifVersion: int = None

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/data/{record_id}"
        query_params = []
        if columns is not None:
            query_params.append("columns=%s" % ",".join(columns))
        if createOnly is not None:
            query_params.append(f"createOnly={createOnly}")
        if ifVersion is not None:
            query_params.append(f"ifVersion={ifVersion}")
        if query_params:
            url_path += "?" + "&".join(query_params)
        headers = {"content-type": "application/json"}
        return self.request("PUT", url_path, headers, payload)

    def upsertRecordWithID(
        self,
        table_name: str,
        record_id: str,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
        columns: list = None,
        ifVersion: int = None,
    ) -> Response:
        """
        Upsert record with ID
        Path: /db/{db_branch_name}/tables/{table_name}/data/{record_id}
        Method: POST

        :param table_name: str The Table name
        :param record_id: str The Record name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.
        :param columns: list = None Column filters
        :param ifVersion: int = None

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/data/{record_id}"
        query_params = []
        if columns is not None:
            query_params.append("columns=%s" % ",".join(columns))
        if ifVersion is not None:
            query_params.append(f"ifVersion={ifVersion}")
        if query_params:
            url_path += "?" + "&".join(query_params)
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def deleteRecord(
        self,
        table_name: str,
        record_id: str,
        db_name: str = None,
        branch_name: str = None,
        columns: list = None,
    ) -> Response:
        """
        Delete record from table
        Path: /db/{db_branch_name}/tables/{table_name}/data/{record_id}
        Method: DELETE

        :param table_name: str The Table name
        :param record_id: str The Record name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.
        :param columns: list = None Column filters

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/data/{record_id}"
        if columns is not None:
            url_path += "?columns=%s" % ",".join(columns)
        return self.request("DELETE", url_path)

    def updateRecordWithID(
        self,
        table_name: str,
        record_id: str,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
        columns: list = None,
        ifVersion: int = None,
    ) -> Response:
        """
        Update record with ID
        Path: /db/{db_branch_name}/tables/{table_name}/data/{record_id}
        Method: PATCH

        :param table_name: str The Table name
        :param record_id: str The Record name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.
        :param columns: list = None Column filters
        :param ifVersion: int = None

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/data/{record_id}"
        query_params = []
        if columns is not None:
            query_params.append("columns=%s" % ",".join(columns))
        if ifVersion is not None:
            query_params.append(f"ifVersion={ifVersion}")
        if query_params:
            url_path += "?" + "&".join(query_params)
        headers = {"content-type": "application/json"}
        return self.request("PATCH", url_path, headers, payload)

    def bulkInsertTableRecords(
        self,
        table_name: str,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
        columns: list = None,
    ) -> Response:
        """
        Bulk insert records
        Path: /db/{db_branch_name}/tables/{table_name}/bulk
        Method: POST

        :param table_name: str The Table name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.
        :param columns: list = None Column filters

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/bulk"
        if columns is not None:
            url_path += "?columns=%s" % ",".join(columns)
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['xata', 'xata.namespaces', 'xata.namespaces.core', 'xata.namespaces.workspace']

package_data = \
{'': ['*']}

install_requires = \
['orjson>=3.8.1,<4.0.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'requests>=2.28.1,<3.0.0']

setup_kwargs = {
    'name': 'xata',
    'version': '0.5.0',
    'description': 'Python client for Xata.io',
    'long_description': '<p align="center">\n  <img width="200" src="https://raw.githubusercontent.com/xataio/company/main/logo/non-resizable/for-light-backgrounds/colored-with-text.png#gh-light-mode-only" />\n  <img width="200" src="https://raw.githubusercontent.com/xataio/company/main/logo/non-resizable/for-dark-backgrounds/colored-with-text.png#gh-dark-mode-only" />\n</p>\n\n# Python SDK for Xata\n\n[![Documentation Status](https://readthedocs.org/projects/xata-py/badge/?version=latest)](https://xata-py.readthedocs.io/en/latest/?badge=latest) [![PyPI version](https://badge.fury.io/py/xata.svg)](https://badge.fury.io/py/xata)\n\nSimple Python client for xata.io databases. Currently work in progress.\n\nXata is a Serverless Database that is as easy to use as a spreadsheet, has the\ndata integrity of PostgreSQL, and the search and analytics functionality of\nElasticsearch.\n\nThe Python SDK uses type annotations and requires Python 3.8 or higher.\n\nTo install, run:\n\n```\npip install xata\n```\n\nTo learn more about Xata, visit [xata.io](https://xata.io).\n\n- Python client documentation: https://xata-py.readthedocs.io\n- API Reference: https://xata-py.readthedocs.io/en/latest/api.html\n',
    'author': 'Xata',
    'author_email': 'support@xata.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from requests import Response, request

from .errors import RateLimitException


class Namespace:
    """
    Parent class for Namespaces
    """

    def __init__(self, client):
        self.client = client

    def get_scope(self) -> str:
        return self.scope

    def is_control_plane(self) -> bool:
        return self.get_scope() == "core"

    def get_base_url(self) -> str:
        if self.is_control_plane():
            return self.base_url
        # Base URL must be build on the fly as the region & workspace Id can change
        return self.base_url.replace(
            "{workspaceId}", self.client.get_config()["workspaceId"]
        ).replace("{regionId}", self.client.get_config()["region"])

    def request(
        self, http_method: str, url_path: str, headers: dict = {}, payload: dict = None
    ) -> Response:
        headers = {
            **headers,
            **self.client.get_headers(),
        }  # TODO use "|" when client py min version >= 3.9
        url = "%s/%s" % (self.get_base_url(), url_path.lstrip("/"))

        if payload is None:
            resp = request(http_method, url, headers=headers)
        else:
            resp = request(http_method, url, headers=headers, json=payload)

        if resp.status_code == 429:
            raise RateLimitException(f"Rate limited: {resp.json()}")

        return resp



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Branch
# Branch management.
# Specification: workspace:v1.0
# Base URL: https://{workspaceId}.{regionId}.xata.sh
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Branch(Namespace):

    base_url = "https://{workspaceId}.{regionId}.xata.sh"
    scope = "workspace"

    def getBranchList(self, db_name: str) -> Response:
        """
        List all available Branches
        Path: /dbs/{db_name}
        Method: GET

        :param db_name: str The Database Name

        :return Response
        """
        url_path = f"/dbs/{db_name}"
        return self.request("GET", url_path)

    def getBranchDetails(
        self, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Get branch schema and metadata
        Path: /db/{db_branch_name}
        Method: GET

        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}"
        return self.request("GET", url_path)

    def createBranch(
        self,
        payload: dict,
        db_name: str = None,
        branch_name: str = None,
        _from: str = None,
    ) -> Response:
        """
        Create Database branch
        Path: /db/{db_branch_name}
        Method: PUT

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.
        :param _from: str = None Name of source branch to branch the new schema from

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}"
        if _from is not None:
            url_path += "?from={_from}"
        headers = {"content-type": "application/json"}
        return self.request("PUT", url_path, headers, payload)

    def deleteBranch(self, db_name: str = None, branch_name: str = None) -> Response:
        """
        Delete the branch in the database and all its resources
        Path: /db/{db_branch_name}
        Method: DELETE

        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}"
        return self.request("DELETE", url_path)

    def getBranchMetadata(
        self, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Get Branch Metadata
        Path: /db/{db_branch_name}/metadata
        Method: GET

        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/metadata"
        return self.request("GET", url_path)

    def updateBranchMetadata(
        self, payload: dict, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Update the branch metadata
        Path: /db/{db_branch_name}/metadata
        Method: PUT

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/metadata"
        headers = {"content-type": "application/json"}
        return self.request("PUT", url_path, headers, payload)

    def getBranchStats(self, db_name: str = None, branch_name: str = None) -> Response:
        """
        Get branch usage metrics.
        Path: /db/{db_branch_name}/stats
        Method: GET

        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/stats"
        return self.request("GET", url_path)

    def getGitBranchesMapping(self, db_name: str) -> Response:
        """
        Lists all the git branches in the mapping, and their associated Xata branches.  Example
        response:  ```json {   "mappings": [       {         "gitBranch": "main",
        "xataBranch": "main"       },       {         "gitBranch": "gitBranch1",
        "xataBranch": "xataBranch1"       }       {         "gitBranch": "xataBranch2",
        "xataBranch": "xataBranch2"       }   ] } ```
        Path: /dbs/{db_name}/gitbranches
        Method: GET

        :param db_name: str The Database Name

        :return Response
        """
        url_path = f"/dbs/{db_name}/gitbranches"
        return self.request("GET", url_path)

    def addGitBranchesEntry(self, db_name: str, payload: dict) -> Response:
        """
        Adds an entry to the mapping of git branches to Xata branches.  The git branch and the
        Xata branch must be present in the body of the request.  If the Xata branch doesn't exist,
        a 400 error is returned.  If the git branch is already present in the mapping, the old
        entry is overwritten, and a warning message is included in the response.  If the git
        branch is added and didn't exist before, the response code is 204. If the git branch
        existed and it was overwritten, the response code is 201.  Example request:  ```json //
        POST https://tutorial-ng7s8c.xata.sh/dbs/demo/gitBranches {   "gitBranch": "fix/bug123",
        "xataBranch": "fix_bug" } ```
        Path: /dbs/{db_name}/gitbranches
        Method: POST

        :param db_name: str The Database Name
        :param payload: dict content

        :return Response
        """
        url_path = f"/dbs/{db_name}/gitbranches"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def removeGitBranchesEntry(self, db_name: str, gitBranch: str) -> Response:
        """
        Removes an entry from the mapping of git branches to Xata branches.  The name of the git
        branch must be passed as a query parameter.  If the git branch is not found, the endpoint
        returns a 404 status code.  Example request:  ```json // DELETE https://tutorial-
        ng7s8c.xata.sh/dbs/demo/gitBranches?gitBranch=fix%2Fbug123 ```
        Path: /dbs/{db_name}/gitbranches
        Method: DELETE

        :param db_name: str The Database Name
        :param gitBranch: str The Git Branch to remove from the mapping

        :return Response
        """
        url_path = f"/dbs/{db_name}/gitbranches"
        if gitBranch is not None:
            url_path += "?gitBranch={gitBranch}"
        return self.request("DELETE", url_path)

    def resolveBranch(
        self, db_name: str, gitBranch: str = None, fallbackBranch: str = None
    ) -> Response:
        """
        In order to resolve the database branch, the following algorithm is used: * if the
        `gitBranch` was provided and is found in the [git branches mapping](/api-
        reference/dbs/db_name/gitBranches), the associated Xata branch is returned * else, if a
        Xata branch with the exact same name as `gitBranch` exists, return it * else, if
        `fallbackBranch` is provided and a branch with that name exists, return it * else, return
        the default branch of the DB (`main` or the first branch)  Example call:  ```json // GET
        https://tutorial-
        ng7s8c.xata.sh/dbs/demo/dbs/demo/resolveBranch?gitBranch=test&fallbackBranch=tsg ```
        Example response:  ```json {   "branch": "main",   "reason": {     "code":
        "DEFAULT_BRANCH",     "message": "Default branch for this database (main)"   } } ```
        Path: /dbs/{db_name}/resolvebranch
        Method: GET

        :param db_name: str The Database Name
        :param gitBranch: str = None The Git Branch
        :param fallbackBranch: str = None Default branch to fallback to

        :return Response
        """
        url_path = f"/dbs/{db_name}/resolvebranch"
        query_params = []
        if gitBranch is not None:
            query_params.append(f"gitBranch={gitBranch}")
        if fallbackBranch is not None:
            query_params.append(f"fallbackBranch={fallbackBranch}")
        if query_params:
            url_path += "?" + "&".join(query_params)
        return self.request("GET", url_path)



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import json
import os
import uuid
from typing import Literal, Optional
from urllib.parse import urljoin

import requests
from dotenv import dotenv_values

from .errors import (
    BadRequestException,
    RateLimitException,
    RecordNotFoundException,
    ServerErrorException,
    UnauthorizedException,
)
from .namespaces.core.authentication import Authentication
from .namespaces.core.databases import Databases
from .namespaces.core.invites import Invites
from .namespaces.core.users import Users
from .namespaces.core.workspaces import Workspaces
from .namespaces.workspace.branch import Branch
from .namespaces.workspace.migrations import Migrations
from .namespaces.workspace.records import Records
from .namespaces.workspace.search_and_filter import Search_and_filter
from .namespaces.workspace.table import Table

# TODO this is a manual task, to keep in sync with pyproject.toml
# could/should be automated to keep in sync
__version__ = "0.5.0"

PERSONAL_API_KEY_LOCATION = "~/.config/xata/key"
DEFAULT_BASE_URL_DOMAIN = "xata.sh"
DEFAULT_CONTROL_PLANE_DOMAIN = "api.xata.io"
DEFAULT_REGION = "us-east-1"
DEFAULT_BRANCH_NAME = "main"
CONFIG_LOCATION = ".xatarc"

ApiKeyLocation = Literal["env", "dotenv", "profile", "parameter"]
WorkspaceIdLocation = Literal["parameter", "env", "config"]


class XataClient:
    """This is the Xata Client. When initialized, it will attempt to read the relevant
    configuration (API key, workspace ID, database name, branch name) from the following
    sources in order:

    * parameters passed to the constructor
    * environment variables
    * .env file
    * .xatarc configuration file

    :meta public:
    :param api_key: API key to use for authentication.
    :param db_url: The database URL to use. If this is specified,
                   then workspace_id, region and db_name must not be specified.
    :param workspace_id: The workspace ID to use.
    :param region: The region to use.
    :param db_name: The database name to use.
    :param branch_name: The branch name to use. Defaults to `main`
    :param base_url_domain: The domain to use for the base URL. Defaults to xata.sh.
    :param control_plane_domain: The domain to use for the control plane. Defaults to api.xata.io.
    """

    configRead: bool = False
    config = None
    namespaces = {}  # lazy loading container for the namespaces

    def __init__(
        self,
        api_key: str = None,
        base_url_domain: str = DEFAULT_BASE_URL_DOMAIN,
        control_plane_domain: str = DEFAULT_CONTROL_PLANE_DOMAIN,
        region: str = DEFAULT_REGION,
        workspace_id: str = None,
        db_name: str = None,
        db_url: str = None,
        branch_name: str = DEFAULT_BRANCH_NAME,
    ):
        """Constructor for the XataClient."""
        if db_url is not None:
            if workspace_id is not None or db_name is not None:
                raise Exception(
                    "Cannot specify both db_url and workspace_id/region/db_name"
                )
            workspace_id, region, db_name = self.parse_database_url(db_url)

        if api_key is None:
            self.api_key, self.api_key_location = self.get_api_key()
        else:
            self.api_key, self.api_key_location = api_key, "parameter"
        if workspace_id is None:
            (
                self.workspace_id,
                self.region,
                self.workspace_id_location,
            ) = self.get_workspace_id()
        else:
            self.workspace_id = workspace_id
            self.workspace_id_location = "parameter"
            self.region = region
        self.base_url = f"https://{self.workspace_id}.{self.region}.{base_url_domain}"
        self.control_plane_url = (
            f"https://{control_plane_domain}/workspaces/{self.workspace_id}/"
        )

        self.db_name = (
            self.get_database_name_if_configured() if db_name is None else db_name
        )
        self.branch_name = (
            self.get_branch_name_if_configured() if branch_name is None else branch_name
        )
        self.headers = {
            "authorization": f"Bearer {self.api_key}",
            "user-agent": f"xataio/xata-py:{__version__}",
            "x-xata-client-id": str(uuid.uuid4()),
            "x-xata-session-id": str(uuid.uuid4()),
            "x-xata-agent": f"client=PY_SDK;version={__version__};",
        }

    def get_config(self) -> dict:
        """
        Get the configuration
        """
        return {
            "apiKey": self.api_key,
            "apiKeyLocation": self.api_key_location,
            "workspaceId": self.workspace_id,
            "region": self.region,
            "dbName": self.db_name,
            "branchName": self.branch_name,
            "version": __version__,
        }

    def get_headers(self) -> dict:
        """
        Get the static headers that are iniatilized on client init.
        """
        return self.headers

    def get_api_key(self) -> tuple[str, ApiKeyLocation]:
        if os.environ.get("XATA_API_KEY") is not None:
            return os.environ.get("XATA_API_KEY"), "env"

        envVals = dotenv_values(".env")
        if envVals.get("XATA_API_KEY") is not None:
            return envVals.get("XATA_API_KEY"), "dotenv"

        if os.path.isfile(os.path.expanduser(PERSONAL_API_KEY_LOCATION)):
            with open(os.path.expanduser(PERSONAL_API_KEY_LOCATION), "r") as f:
                return f.read().strip(), "profile"

        raise Exception(
            f"No API key found. Searched in `XATA_API_KEY` env, "
            f"`{PERSONAL_API_KEY_LOCATION}`, and `{os.path.abspath('.env')}`"
        )

    def get_workspace_id(self) -> tuple[str, str, WorkspaceIdLocation]:
        if os.environ.get("XATA_WORKSPACE_ID") is not None:
            return (
                os.environ.get("XATA_WORKSPACE_ID"),
                os.environ.get("XATA_REGION", DEFAULT_REGION),
                "env",
            )

        envVals = dotenv_values(".env")
        if envVals.get("XATA_WORKSPACE_ID") is not None:
            return (
                envVals.get("XATA_WORKSPACE_ID"),
                envVals.get("XATA_REGION", DEFAULT_REGION),
                "dotenv",
            )

        self.ensure_config_read()
        if self.config is not None and self.config.get("databaseURL"):
            workspaceID, region, _ = self.parse_database_url(
                self.config.get("databaseURL")
            )
            return workspaceID, region, "config"
        raise Exception(
            f"No workspace ID found. Searched in `XATA_WORKSPACE_ID` env, "
            f"`{PERSONAL_API_KEY_LOCATION}`, and `{os.path.abspath('.env')}`"
        )

    def get_database_name_if_configured(self) -> str:
        self.ensure_config_read()
        if self.config is not None and self.config.get("databaseURL"):
            _, _, db_name = self.parse_database_url(self.config.get("databaseURL"))
            return db_name
        return None

    def get_branch_name_if_configured(self) -> str:
        # TODO: resolve branch name by the current git branch
        return os.environ.get("XATA_BRANCH")

    def get_db_branch_name(self, db_name: str = None, branch_name: str = None) -> str:
        """
        Get Database with branch name, format: {db_name}:{branch_name}
        The name can be build with passed params or from config or a combination of both

        :param db_name: str
        :branch_name: str

        :return str
        """
        if db_name is None:
            db_name = self.db_name
        if branch_name is None:
            branch_name = self.branch_name
        return f"{db_name}:{branch_name}"

    def request(self, method, urlPath, cp=False, headers={}, expect_codes=[], **kwargs):
        headers = {
            **headers,
            **self.headers,
        }  # TODO use "|" when client py min version >= 3.9

        base_url = self.base_url if not cp else self.control_plane_url
        url = urljoin(base_url, urlPath.lstrip("/"))

        resp = requests.request(method, url, headers=headers, **kwargs)
        if resp.status_code > 299:
            if resp.status_code in expect_codes:
                return resp
            if resp.status_code == 401:
                raise UnauthorizedException(
                    f"Unauthorized: {resp.json()} API key location: {self.api_key_location}"
                )
            elif resp.status_code == 429:
                raise RateLimitException(f"Rate limited: {resp.json()}")
            elif resp.status_code >= 399 and resp.status_code < 500:
                raise BadRequestException(resp.status_code, resp.json().get("message"))
            elif resp.status_code >= 500:
                raise ServerErrorException(f"Server error: {resp.text}")
            raise Exception(f"{resp.status_code} {resp.text}")
        return resp

    def ensure_config_read(self) -> bool:
        if self.configRead:
            return False
        if os.path.isfile(CONFIG_LOCATION):
            with open(CONFIG_LOCATION, "r") as f:
                self.config = json.load(f)
        self.configRead = True
        return True

    def parse_database_url(self, databaseURL: str) -> tuple[str, str, str]:
        (_, _, host, _, db) = databaseURL.split("/")
        if host == "":
            raise Exception("Invalid database URL")
        parts = host.split(".")
        workspaceId = parts[0]
        region = parts[1]
        return workspaceId, region, db

    def get(self, urlPath, headers={}, **kwargs):
        """Send a GET request to the Xata API. This is a wrapper around
        the `requests` library and accepts the same parameters as the `get`
        method of the `requests` library.
        """

        return self.request("GET", urlPath, headers=headers, **kwargs)

    def post(self, urlPath, headers={}, **kwargs):
        """Send a POST request to the Xata API. This is a wrapper around
        the `requests` library and accepts the same parameters as the `post`
        method of the `requests` library.
        """
        return self.request("POST", urlPath, headers=headers, **kwargs)

    def put(self, urlPath, headers={}, **kwargs):
        """Send a PUT request to the Xata API. This is a wrapper around
        the `requests` library and accepts the same parameters as the `put`
        method of the `requests` library.
        """
        return self.request("PUT", urlPath, headers=headers, **kwargs)

    def delete(self, urlPath, headers={}, **kwargs):
        """Send a DELETE request to the Xata API. This is a wrapper around
        the `requests` library and accepts the same parameters as the `delete`
        method of the `requests` library.
        """
        return self.request("DELETE", urlPath, headers=headers, **kwargs)

    def patch(self, urlPath, headers={}, **kwargs):
        """Send a PATCH request to the Xata API. This is a wrapper around
        the `requests` library and accepts the same parameters as the `patch`
        method of the `requests` library.
        """
        return self.request("PATCH", urlPath, headers=headers, **kwargs)

    def request_body_from_params(
        self,
        columns: list[str] = None,
        filter: dict = None,
        sort: dict = None,
        page: dict = None,
    ) -> dict:

        body = {}
        if columns is not None:
            body["columns"] = columns
        if filter is not None:
            body["filter"] = filter
        if sort is not None:
            body["sort"] = sort
        if page is not None:
            body["page"] = page
        return body

    def db_and_branch_names_from_params(self, db_name, branch_name) -> tuple[str, str]:
        db_name = db_name or self.db_name
        branch_name = branch_name or self.branch_name
        if db_name is None:
            raise Exception(
                "Database name is not configured. Please set it via `xata init` or pass it as a parameter."
            )
        if branch_name is None:
            raise Exception(
                "Branch name is not configured. Please set it in the `XATA_BRANCH` env var or pass it as a parameter."
            )
        return db_name, branch_name

    def set_db_and_branch_names(self, db_name: str = None, branch_name: str = None):
        if db_name is not None:
            self.db_name = db_name
        if branch_name is not None:
            self.branch_name = branch_name

    def query(
        self,
        table: str,
        db_name: str = None,
        branch_name: str = None,
        columns: list[str] = None,
        filter: dict = None,
        sort: dict = None,
        page: dict = None,
    ) -> dict:
        """Query a table.

        :meta public:
        :param table: The name of the table to query.
        :param db_name: The name of the database to query. If not provided, the database name
                    from the client obejct is used.
        :param branch_name: The name of the branch to query. If not provided, the branch name
                        from the client obejct is used.
        :param columns: A list of column names to return. If not provided, all columns are returned.
        :param filter: A filter expression to apply to the query.
        :param sort: A sort expression to apply to the query.
        :param page: A page expression to apply to the query.
        :return: A page of results.
        """

        db_name, branch_name = self.db_and_branch_names_from_params(
            db_name, branch_name
        )
        body = self.request_body_from_params(columns, filter, sort, page)
        result = self.post(
            f"/db/{db_name}:{branch_name}/tables/{table}/query", json=body
        )
        return result.json()

    def get_first(
        self,
        table: str,
        db_name: str = None,
        branch_name: str = None,
        columns: list[str] = None,
        filter: dict = None,
        sort: dict = None,
    ) -> dict:
        """Get the first record from a table respecting the provided filters and sort order.

        :meta public:
        :param table: The name of the table to query.
        :param db_name: The name of the database to query. If not provided, the database name
                        from the client obejct is used.
        :param branch_name: The name of the branch to query. If not provided, the branch name
                        from the client obejct is used.
        :param columns: A list of column names to return. If not provided, all columns are returned.
        :param filter: A filter expression to apply to the query.
        :param sort: A sort expression to apply to the query.
        :return: A record as a dictionary.
        """

        page = {"size": 1}
        db_name, branch_name = self.db_and_branch_names_from_params(
            db_name, branch_name
        )
        body = self.request_body_from_params(columns, filter, sort, page)
        result = self.post(
            f"/db/{db_name}:{branch_name}/tables/{table}/query", json=body
        )
        data = result.json()
        if len(data.get("records", [])) == 0:
            return None
        return data.get("records")[0]

    def get_by_id(
        self,
        table: str,
        id: str,
        db_name: str = None,
        branch_name: str = None,
    ) -> Optional[dict]:
        """Get a specific record by its ID. Returns None if an record with that ID
        doesn't exist.

        :meta public:
        :param table: The name of the table to query.
        :param id: The ID of the record to get.
        :param db_name: The name of the database to query. If not provided, the database name
                        from the client obejct is used.
        :param branch_name: The name of the branch to query. If not provided, the branch name
                        from the client obejct is used.
        :return: A record as a dictionary or None if it doesn't exist.
        """
        db_name, branch_name = self.db_and_branch_names_from_params(
            db_name, branch_name
        )
        result = self.get(
            f"/db/{db_name}:{branch_name}/tables/{table}/data/{id}", expect_codes=[404]
        )
        if result.status_code == 404:
            return None
        return result.json()

    def create(
        self,
        table: str,
        record: dict,
        id: str = None,
        db_name: str = None,
        branch_name: str = None,
    ) -> str:
        """Create a record in a table. If an ID is not provided, one will be generated.
        If the ID is provided and a record with that ID already exists, an error is returned.

        :meta public:
        :param table: The name of the table to query.
        :param id: The ID of the record to create. If not provided, one will be generated.
        :param record: The record to create, as dict.
        :param db_name: The name of the database to query. If not provided, the database name
                        from the client obejct is used.
        :param branch_name: The name of the branch to query. If not provided, the branch name
                        from the client obejct is used.
        :return: The ID of the created record.
        """

        db_name, branch_name = self.db_and_branch_names_from_params(
            db_name, branch_name
        )
        if id is not None:
            self.put(
                f"/db/{db_name}:{branch_name}/tables/{table}/data/{id}",
                params=dict(createOnly=True),
                json=record,
            )
            return id

        result = self.post(
            f"/db/{db_name}:{branch_name}/tables/{table}/data", json=record
        )
        return result.json()["id"]

    def create_or_update(
        self,
        table: str,
        id: str,
        record: dict,
        db_name: str = None,
        branch_name: str = None,
    ) -> str:
        """Create or updated a record in a table. If a record with the same id already
        exists, it will be updated. Only the provided columns in record are replaced, is
        a column is not present explicitely in record, it is not updated.

        :meta public:
        :param table: The name of the table to query.
        :param id: The ID of the record to create or update.
        :param record: The record to create or the keys/values to use to update, as dict.
        :param db_name: The name of the database to query. If not provided, the database name
                        from the client obejct is used.
        :param branch_name: The name of the branch to query. If not provided, the branch name
                        from the client obejct is used.
        :return: The ID of the created or updated record.
        """
        db_name, branch_name = self.db_and_branch_names_from_params(
            db_name, branch_name
        )
        result = self.post(
            f"/db/{db_name}:{branch_name}/tables/{table}/data/{id}", json=record
        )
        return result.json()["id"]

    def create_or_replace(
        self,
        table: str,
        id: str,
        record: dict,
        db_name: str = None,
        branch_name: str = None,
    ) -> str:
        """Create or replace a record in a table. If a record with the same id already
        exists, it will be relaced completely.

        :meta public:
        :param table: The name of the table to query.
        :param id: The ID of the record to create or replace.
        :param record: The record to create or replace, as dict.
        :param db_name: The name of the database to query. If not provided, the database name
                        from the client obejct is used.
        :param branch_name: The name of the branch to query. If not provided, the branch name
                        from the client obejct is used.
        :return: The ID of the created or updated record.
        """
        db_name, branch_name = self.db_and_branch_names_from_params(
            db_name, branch_name
        )
        result = self.put(
            f"/db/{db_name}:{branch_name}/tables/{table}/data/{id}", json=record
        )
        return result.json()["id"]

    def update(
        self,
        table: str,
        id: str,
        record: dict,
        ifVersion: Optional[int] = None,
        db_name: str = None,
        branch_name: str = None,
    ) -> Optional[dict]:
        """Updates the record with the given key-value pairs in the record. The columns
        that aren't explicitely provided are left unchanged. If no record with the given
        id exists, None is returned. If the ifVersion condition is not respected, None is
        returned.

        :meta public:
        :param table: The name of the table to query.
        :param id: The ID of the record to update.
        :param record: The key-value pairs to update.
        :param ifVersion: Only perform the update if the version of the record matches this value.
        :param db_name: The name of the database to query. If not provided, the database name
                        from the client obejct is used.
        :param branch_name: The name of the branch to query. If not provided, the branch name
                        from the client obejct is used.
        :return: The updated record.
        """
        db_name, branch_name = self.db_and_branch_names_from_params(
            db_name, branch_name
        )
        params = {"columns": "*"}
        if ifVersion is not None:
            params["ifVersion"] = str(ifVersion)
        result = self.patch(
            f"/db/{db_name}:{branch_name}/tables/{table}/data/{id}",
            params=params,
            json=record,
            expect_codes=[422, 404],
        )
        if result.status_code == 404:
            return (
                None  # TODO: I would prefer to raise here, but there is a backend issue
            )
        if result.status_code == 422:
            return None
        return result.json()

    def delete_record(
        self, table: str, id: str, db_name: str = None, branch_name: str = None
    ) -> Optional[dict]:
        """Deletes the record with the given ID. Returns the record as it was just before
        deletion. If no record with the given ID exists, raises RecordNotFoundException.

        :meta public:
        :param table: The name of the table to query.
        :param id: The ID of the record to delete.
        :param db_name: The name of the database to query. If not provided, the database name
                        from the client obejct is used.
        :param branch_name: The name of the branch to query. If not provided, the branch name
                        from the client obejct is used.
        :return: The deleted record.
        """
        db_name, branch_name = self.db_and_branch_names_from_params(
            db_name, branch_name
        )
        params = {"columns": "*"}
        result = self.delete(
            f"/db/{db_name}:{branch_name}/tables/{table}/data/{id}",
            params=params,
            expect_codes=[404],
        )
        if result.status_code == 404:
            raise RecordNotFoundException(id)
        return result.json()

    def search(
        self,
        query: str,
        query_params: dict = {},
        db_name: str = None,
        branch_name: str = None,
    ) -> Optional[dict]:
        """This endpoint performs full text search across an entire database branch.

        API docs: https://xata.io/docs/api-reference/db/db_branch_name/search#free-text-search

        :meta public:
        :param query: search string
        :param query_params: more granular search criteria, see API docs for options
        :param db_name: The name of the database to query. If not provided, the database name
                        from the client obejct is used.
        :param branch_name: The name of the branch to query. If not provided, the branch name
                        from the client obejct is used.

        :return set of matching records
        """
        db_name, branch_name = self.db_and_branch_names_from_params(
            db_name, branch_name
        )
        query_params["query"] = query.strip()
        result = self.post(
            f"/db/{db_name}:{branch_name}/search",
            json=query_params,
            expect_codes=[200, 400, 403, 404, 500],
        )

        if result.status_code == 400:
            raise BadRequestException(result.status_code, result.json()["message"])
        return result.json()

    def search_table(
        self,
        table_name: str,
        query: str,
        query_params: dict = {},
        db_name: str = None,
        branch_name: str = None,
    ) -> Optional[dict]:
        """Run a free text search operation in a particular table.

        :meta public:
        :param table_name: table to search
        :param query: search string
        :param query_params: more granular search criteria, see API docs for options
        :param db_name: The name of the database to query. If not provided, the database name
                        from the client obejct is used.
        :param branch_name: The name of the branch to query. If not provided, the branch name
                        from the client obejct is used.

        :return set of matching records
        """
        db_name, branch_name = self.db_and_branch_names_from_params(
            db_name, branch_name
        )
        query_params["query"] = query.strip()
        result = self.post(
            f"/db/{db_name}:{branch_name}/tables/{table_name}/search",
            json=query_params,
            expect_codes=[200, 400, 403, 404, 500],
        )

        if result.status_code == 400:
            raise BadRequestException(result.status_code, result.json()["message"])
        return result.json()

    # --------------------------------------------------- #
    #
    # Namespace: CORE
    #
    # --------------------------------------------------- #

    def authentication(self) -> Authentication:
        """
        Authentication Namespace
        scope: core
        :return Authentication
        """
        if "authentication" not in self.namespaces:
            self.namespaces["authentication"] = Authentication(self)
        return self.namespaces["authentication"]

    def databases(self) -> Databases:
        """
        Databases Namespace
        scope: core
        :return Databases
        """
        if "databases" not in self.namespaces:
            self.namespaces["databases"] = Databases(self)
        return self.namespaces["databases"]

    def invites(self) -> Invites:
        """
        Invites Namespace
        scope: core
        :return Invites
        """
        if "invites" not in self.namespaces:
            self.namespaces["invites"] = Invites(self)
        return self.namespaces["invites"]

    def users(self) -> Users:
        """
        Users Namespace
        scope: core
        :return Users
        """
        if "users" not in self.namespaces:
            self.namespaces["users"] = Users(self)
        return self.namespaces["users"]

    def workspaces(self) -> Workspaces:
        """
        Workspaces Namespace
        scope: core
        :return Workspaces
        """
        if "workspaces" not in self.namespaces:
            self.namespaces["workspaces"] = Workspaces(self)
        return self.namespaces["workspaces"]

    # --------------------------------------------------- #
    #
    # Namespace: WORKSPACE
    #
    # --------------------------------------------------- #

    def branch(self) -> Branch:
        """
        Branch Namespace
        scope: workspace
        :return Branch
        """
        if "branch" not in self.namespaces:
            self.namespaces["branch"] = Branch(self)
        return self.namespaces["branch"]

    def migrations(self) -> Migrations:
        """
        Migrations Namespace
        scope: workspace
        :return Migrations
        """
        if "migrations" not in self.namespaces:
            self.namespaces["migrations"] = Migrations(self)
        return self.namespaces["migrations"]

    def records(self) -> Records:
        """
        Records Namespace
        scope: workspace
        :return Records
        """
        if "records" not in self.namespaces:
            self.namespaces["records"] = Records(self)
        return self.namespaces["records"]

    def search_and_filter(self) -> Search_and_filter:
        """
        Search_and_Filter Namespace
        scope: workspace
        :return Search_and_filter
        """
        if "search_and_filter" not in self.namespaces:
            self.namespaces["search_and_filter"] = Search_and_filter(self)
        return self.namespaces["search_and_filter"]

    def table(self) -> Table:
        """
        Table Namespace
        scope: workspace
        :return Table
        """
        if "table" not in self.namespaces:
            self.namespaces["table"] = Table(self)
        return self.namespaces["table"]



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Migrations
# Branch schema migrations and history.
# Specification: workspace:v1.0
# Base URL: https://{workspaceId}.{regionId}.xata.sh
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Migrations(Namespace):

    base_url = "https://{workspaceId}.{regionId}.xata.sh"
    scope = "workspace"

    def getBranchMigrationHistory(
        self, payload: dict, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Get branch migration history [deprecated]
        Path: /db/{db_branch_name}/migrations
        Method: GET

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/migrations"
        headers = {"content-type": "application/json"}
        return self.request("GET", url_path, headers, payload)

    def getBranchMigrationPlan(
        self, payload: dict, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Compute a migration plan from a target schema the branch should be migrated too.
        Path: /db/{db_branch_name}/migrations/plan
        Method: POST

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/migrations/plan"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def executeBranchMigrationPlan(
        self, payload: dict, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Apply a migration plan to the branch
        Path: /db/{db_branch_name}/migrations/execute
        Method: POST

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/migrations/execute"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def getBranchSchemaHistory(
        self, payload: dict, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Query schema history.
        Path: /db/{db_branch_name}/schema/history
        Method: POST

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/schema/history"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def compareBranchWithUserSchema(
        self, payload: dict, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Compare branch with user schema.
        Path: /db/{db_branch_name}/schema/compare
        Method: POST

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/schema/compare"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def compareBranchSchemas(
        self, payload: dict, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Compare branch schemas.
        Path: /db/{db_branch_name}/schema/compare/{branch_name}
        Method: POST

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/schema/compare/{branch_name}"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def updateBranchSchema(
        self, payload: dict, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Update Branch schema
        Path: /db/{db_branch_name}/schema/update
        Method: POST

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/schema/update"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def previewBranchSchemaEdit(
        self, payload: dict, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Preview branch schema edits.
        Path: /db/{db_branch_name}/schema/preview
        Method: POST

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/schema/preview"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def applyBranchSchemaEdit(
        self, payload: dict, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Apply edit script.
        Path: /db/{db_branch_name}/schema/apply
        Method: POST

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/schema/apply"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)
